#ifndef	ASN1FIX_PARAMETERIZATION_H
#define	ASN1FIX_PARAMETERIZATION_H

asn1p_expr_t *asn1f_parameterization_fork(arg_t *arg, asn1p_expr_t *expr, asn1p_expr_t *rhs_pspecs);

#endif	/* ASN1FIX_PARAMETERIZATION_H */
